{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE OverloadedLabels #-}
--{-# LANGUAGE FlexibleContexts #-}

import Control.Exception (bracket)
import Control.Monad (void)
import qualified Data.Text as Text
import qualified Data.Text.IO as Text
import           Data.Text (Text)
import           Data.Maybe (fromJust)
import GHC.Generics
import qualified GI.Gdk as Gdk
import qualified GI.GdkX11 as GdkX11
import qualified GI.Gtk as Gtk
import qualified GI.GLib as GLib
import qualified Graphics.X11 as X
import qualified GI.Pango.Enums as Pango
import Data.GI.Base
-- TODO: is Regex.TDFA good / fast?
import Text.Regex.TDFA
import Text.Regex.TDFA.Text

import Network.Lean.Client
import Network.Lean.Lai (Request(..))
import Data.Serialize.Ldsf.Types
import qualified Data.Serialize.Ldsf.Lean.InterfaceFunction as IF
import qualified Data.Serialize.Ldsf.Lean.Category.Browser as CategoryBrowser
import qualified Data.Serialize.Ldsf.Lean.Category.Text as CategoryText
import qualified Data.Text as T
import qualified Data.Text.Lazy as TL

import System.Environment (getArgs)

-- given a text, return ascii text
-- TODO: FIXME: NOT YET IMPLEMENTED
sanitize :: Text -> Text
sanitize txt = txt

-- given a text, find the beginning and end of all links
-- TODO: FIXME: no property tests available
parse :: Text -> [(Int, Int)]
parse txt = map (\(a, b) -> (a, a + b)) matches
    where
        matches = getAllMatches (txt =~ ("lean://[!*'();:@&=+$,/A-Za-z0-9\\-_\\\\.~]+" :: Text))

-- | TODO: FIXME: generalize, merge with 'parse'.
parseUrl :: String -> (String, String, String)
parseUrl url = (hostname, category, file)
    where
        (_, _, _, [hostname, category, file]) = (url =~ ("lean://(.+)/(.+)/(.+)" :: String)) :: (String, String, String, [String])

setTextViewCursor :: Gtk.TextView -> Maybe Gdk.Cursor -> IO (Either Text ())
setTextViewCursor textView cursor = do
    maybeGdkWin <- Gtk.textViewGetWindow textView Gtk.TextWindowTypeText
    case maybeGdkWin of
         Just gdkWin -> do
             Gdk.windowSetCursor gdkWin cursor
             pure $ Right ()
         Nothing  -> pure $ Left "Failed to retrieve GDK window from textview."         

textGet :: String -> String -> String -> IO (Either String Text)
textGet localport targetHostname targetFile = do
    bracket (createClient "127.0.0.1" localport)
            (destroyClient)
            $ \client -> do
                -- request resource from server
                let f1 = CategoryText.functionGet
                    r1 = Request
                        { requestHostname = T.pack targetHostname
                        , requestCategory = IF.category f1
                        , requestVersion = IF.version f1
                        , requestFunction = IF.id f1
                        , requestArguments = IF.putArgs f1 [ValueText8 $ T.pack targetFile]
                        , requestInternal = Nothing
                        }
                    getResult1 = IF.getResult f1
                maybeResponse <- request client r1 getResult1
                case maybeResponse of
                    Nothing                 -> pure $ Left "failed r1"
                    Just (header, response) -> do
                        putStrLn $ "server responded:" ++ show response
                        pure $ case responsePayload response of
                            [ValueText32 v] -> Right $ sanitize $ TL.toStrict v
                            _               -> Left "failed...."

getLink :: Int -> [(Int, Int)] -> (Int, Int)
getLink offset parsed = head xs -- TODO: FIXME: avoid partial, do something smarter
    where
        xs = filter (\(a, b) -> offset >= a && offset <= b) parsed

setView :: (Gtk.IsTextTag a, Gtk.IsTextBuffer b) => a -> b -> Either String Text -> IO ()
setView linkTag textBuffer (Right txt) = do
    let parsed = parse txt
    
    Gtk.textBufferSetText textBuffer txt (toEnum $ Text.length txt)
    
    -- apply tag formatting
    let f (s, e) = do
            start <- Gtk.textBufferGetIterAtOffset textBuffer (toEnum s)
            end <- Gtk.textBufferGetIterAtOffset textBuffer (toEnum e)
            pure (start, end)
        a (s, e) = Gtk.textBufferApplyTag textBuffer linkTag s e
    iters <- mapM f parsed
    mapM_ a iters
setView _ _ (Left msg) = fail msg

main :: IO ()
main = do
    -- TODO: FIXME: error handling, don't use (!!)
    args <- getArgs
    let localport = args !! 0
        url = args !! 1
        (urlHostname, urlCategory, urlFile) = parseUrl url
    
    result <- textGet localport urlHostname urlFile
    
    Gtk.init Nothing
    
    maybeDisplay <- Gdk.displayGetDefault
    display <- case maybeDisplay of
        Just d  -> pure d
        Nothing -> error "Failed to open diplay."
                      
    cursorDefault <- Gdk.cursorNewFromName display "default"
    cursorPointer <- Gdk.cursorNewFromName display "pointer"

    win <- new Gtk.Window
        [ #title := "lean-handler-text"
        , #type := Gtk.WindowTypePopup
        --, #skip-taskbar-hint := True
        --, #skip-pager-hint := True
        ]
    
    on win #destroy Gtk.mainQuit
    
    textView <- new Gtk.TextView
        [ #editable := False
        , #cursorVisible := False
        ]
    textViewBuffer <- textView `get` #buffer
    tagTable <- textViewBuffer `get` #tagTable
    
    linkColor <- new Gdk.RGBA
        [ #red   := 0
        , #green := 0
        , #blue  := 238
        , #alpha := 1.0
        ]
    linkTag <- new Gtk.TextTag
        [ #name := "test"
        , #foregroundRgba := linkColor
        , #underline := Pango.UnderlineSingle
        ]

    on textView #event $ \e -> do
        (extractedCoords, x, y) <- Gdk.eventGetCoords e
        if extractedCoords
            then do
                 void $ setTextViewCursor textView cursorDefault
                 pure False
           else pure True

    on linkTag #event $ \_ ev iter -> do
            (extracted, buttonNr) <- Gdk.eventGetButton ev
            (extractedCoords, x, y) <- Gdk.eventGetCoords ev
            if extracted
                then do
                    if buttonNr == 1 -- TODO: FIXME: don't use magic
                        then do
                            -- iter is where the user clicked, not the start of the tag!
                            iterStart <- Gtk.textBufferGetStartIter textViewBuffer
                            iterEnd <- Gtk.textBufferGetEndIter textViewBuffer
                            txt <- Gtk.textBufferGetText textViewBuffer iterStart iterEnd True
                            offset <- Gtk.textIterGetOffset iter
                            let (start, end) = getLink (fromEnum offset) (parse txt)
                            linkStart <- Gtk.textBufferGetIterAtOffset textViewBuffer (toEnum start)
                            linkEnd <- Gtk.textBufferGetIterAtOffset textViewBuffer (toEnum end)
                            link <- Gtk.textBufferGetSlice textViewBuffer linkStart linkEnd False
                            
                            -- parse url
                            -- TODO: FIXME: error handling
                            let (hostname, category, file) = parseUrl (T.unpack link)
                            result <- textGet localport hostname file
                            setView linkTag textViewBuffer result
                            
                            pure False
                        else pure False
                else if extractedCoords
                    then do
                        void $ setTextViewCursor textView cursorPointer
                        pure False
                    else pure True
    
    Gtk.textTagTableAdd tagTable linkTag
    setView linkTag textViewBuffer result
    
    #add win textView
    
    Gtk.windowSetDecorated win False
    
    #showAll win
    
    -- request embedding
    
    GLib.idleAdd GLib.PRIORITY_HIGH_IDLE $ do
        gdkWindow <- fromJust <$> Gtk.widgetGetWindow win
        x11Window <- Gtk.unsafeCastTo GdkX11.X11Window gdkWindow
        xid_ours <- fromIntegral <$> GdkX11.x11WindowGetXid x11Window
        
        bracket (createClient "127.0.0.1" localport)
                (destroyClient)
                $ \client -> do
                    let f1 = CategoryBrowser.functionEmbed
                        r1 = Request
                            { requestHostname = T.pack ""
                            , requestCategory = IF.category f1
                            , requestVersion = IF.version f1
                            , requestFunction = IF.id f1
                            , requestArguments = IF.putArgs f1 [ValueUInt8 1, ValueUInt32 xid_ours]
                            , requestInternal = Nothing
                            }
                        getResult1 = IF.getResult f1
                    maybeResponse <- request client r1 getResult1
                    case maybeResponse of
                        Nothing                 -> pure $ Left "failed r1"
                        Just (header, response) -> do
                            putStrLn $ "server responded:" ++ show response
                            pure $ case responsePayload response of
                                [ValueText32 v] -> Right $ sanitize $ TL.toStrict v
                                _               -> Left "failed...."
        -- let us hope this works
--         _ <- GLib.timeoutAddSeconds GLib.PRIORITY_DEFAULT 10 $ do
--             Gtk.windowSetDecorated win False
--             pure False
        pure False
    -- embedding

    Gtk.main
