{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE OverloadedLabels #-}

import Control.Concurrent
import Control.Monad (void)
import qualified GI.Gdk as Gdk
import qualified GI.GdkX11 as GdkX11
import qualified GI.Gtk as Gtk
import qualified GI.GLib as GLib
import qualified Graphics.X11 as X
import Data.GI.Base
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Serialize.Ldsf.Types
import qualified Data.Serialize.Ldsf.Lean.InterfaceFunction as IF
import qualified Data.Serialize.Ldsf.Lean.Category.Browser as CategoryBrowser
import qualified Data.Serialize.Ldsf.Lean.Category.Text as CategoryText
import qualified Data.Text as T
import qualified Data.Text.Lazy as TL
import Network.Lean.Lai
import Network.Lean.Lai.Internal (ResponseReceived(..))
import Network.Lean.Lai.Handler.Larp (LarpApplication, run, larpRequestSocketAddr)
import Network.Lean.Types
import Network.Simple.TCP (SockAddr(..))
import System.Process (shell, spawnProcess, spawnCommand)

-- TODO: remove me ?
import Data.Word
import Data.Maybe (fromJust)
import Data.IORef
import Data.List


data Ui = Ui
    { uiReparent :: Word32 -> IO ()
    }
data BrowserState = BrowserState
    { browserUi :: Ui
    }

-- | Reparent a window.
reparentWindow :: Word32 -> Word32 -> IO ()
reparentWindow from to = do
    xd <- X.openDisplay []
    X.reparentWindow xd from to 0 0
    X.flush xd
    X.closeDisplay xd

resizeWindow :: Word32 -> Word32 -> Word32 -> IO ()
resizeWindow xid width height = do
    xd <- X.openDisplay []
    X.resizeWindow xd xid width height
    X.flush xd
    X.closeDisplay xd

buildNotebook = do
    box <- new Gtk.VBox []
    
    label <- new Gtk.Label [ #label := "New Tab" ]
    labelPlus <- new Gtk.Label [ #label := "+" ]
    
    notebook <- new Gtk.Notebook []
    _ <- Gtk.notebookAppendPage notebook box (Just label)
    _ <- Gtk.notebookAppendPage notebook box (Just labelPlus)
    
    pure notebook

buildToolbar = do
    back <- new Gtk.Button [ #label := "⇽" ]
    forward <- new Gtk.Button [ #label := "→" ]
    refresh <- new Gtk.Button  [ #label := "↻" ]
    entry <- new Gtk.Entry []
    settings <- new Gtk.Button [ #label := "☰" ]
    
    box <- new Gtk.HBox []
    
    #packStart box back False True 0
    #packStart box forward False True 0
    #packStart box refresh False True 0
    #packStart box entry True True 0
    #packStart box settings False True 0
    
    pure (box, back, forward, refresh, entry, settings)

buildScrolledWindow = new Gtk.ScrolledWindow [ #vexpand := True ] -- vexpand needed?

-- main entry point and gui handling
main :: IO ()
main = do
    -- set up and handle gui
    Gtk.init Nothing

    win <- new Gtk.Window [ #title := "lean" ]

    on win #destroy Gtk.mainQuit
    
    -- build main ui elements
    notebook <- buildNotebook
    (toolbar, _, _, _, toolbarEntry, _) <- buildToolbar
    scrolledWindow <- buildScrolledWindow
    
    -- pack them in a vertical box
    box <- new Gtk.VBox []
    #packStart box notebook False True 0
    #packStart box toolbar False True 0
    #packStart box scrolledWindow True True 0
    
    -- add them to the window
    #add win box
    
    -- connect signal handlers
    on toolbarEntry #activate $ do
        url <- Gtk.entryGetText toolbarEntry
        launchHandler "lean-handler-text" "9911" (T.unpack url)
        pure ()
    
    -- show all ui elements
    #showAll win
    
    -- removeme
--     let xid_child = 0x4600007
--     on toolbarEntry #activate $ do
--         gdkWindow <- fromJust <$> Gtk.widgetGetWindow scrolledWindow
--         x11Window <- Gtk.unsafeCastTo GdkX11.X11Window gdkWindow
--         xid_ours <- fromIntegral <$> GdkX11.x11WindowGetXid x11Window
--         reparentWindow xid_child xid_ours -- TODO: WINAPI??
--         
--         -- initial resize
--         --width <- Gtk.widgetGetAllocatedWidth scrolledWindow
--         --height <- Gtk.widgetGetAllocatedHeight scrolledWindow
--         --resizeWindow xid_child (fromIntegral width) (fromIntegral height)
--         
--         pure ()
--     -- removeme
    
    refChild <- newIORef 0 :: IO (IORef Word32)
    
    on win #configureEvent $ \e -> do
        -- todo: take from event instead
        putStrLn "configure event"
        windowId <- readIORef refChild
        if windowId == 0
            then pure False
            else do
                width <- Gtk.widgetGetAllocatedWidth scrolledWindow
                height <- Gtk.widgetGetAllocatedHeight scrolledWindow
                resizeWindow windowId (fromIntegral width) (fromIntegral height)
                pure False

    -- functions which we pass to other threads for thread-safe communication
    -- with this thread.
    let guiReparentWindow :: Word32 -> IO ()
        guiReparentWindow windowId = void $ GLib.idleAdd GLib.PRIORITY_HIGH_IDLE $ do
            -- todo fixme
            gdkWindow <- fromJust <$> Gtk.widgetGetWindow scrolledWindow
            x11Window <- Gtk.unsafeCastTo GdkX11.X11Window gdkWindow
            xid_ours <- fromIntegral <$> GdkX11.x11WindowGetXid x11Window
            reparentWindow windowId xid_ours
            
            writeIORef refChild windowId
            
            pure False
    
    -- run networking thread in parallel
    _ <- forkIO $ runNetworking $
        BrowserState
        { browserUi = 
            Ui 
            { uiReparent = guiReparentWindow 
            }
        }
    
    -- run the main loop
    Gtk.main

--data HandlerState = HandlerState { handlerStateHost :: String } deriving (Show)
--type NetworkingState = MVar (Map SockAddr HandlerState)

-- | Run networking thread with empty state.
runNetworking :: BrowserState -> IO ()
runNetworking browser = do
    run 9911 (networking browser)

networking :: BrowserState -> LarpApplication
networking browser request respond = do
    let funcId = requestFunction request
    putStrLn $ "Client request: " ++ show request
    case requestCategory request of
        Browser -> case () of
            () | IF.id CategoryBrowser.functionOpen == funcId  -> handleBrowserOpen browser request respond
               | IF.id CategoryBrowser.functionEmbed == funcId -> handleBrowserEmbed browser request respond
               | otherwise                                     -> handleFunctionNotFound request respond

        Text    -> case () of
            () | IF.id CategoryText.functionGet == funcId -> handleTextGet request respond
               | otherwise                                -> handleFunctionNotFound request respond

        _       -> handleFunctionNotImplemented request respond

{-# WARNING handleBrowserOpen "not implemented" #-}
handleBrowserOpen :: BrowserState -> LarpApplication
handleBrowserOpen browser request respond =
    respond $ Response
        { responseType = Ok
        , responsePayload = [ValueNothing]
        , responseInternal = Nothing
        }

handleBrowserEmbed :: BrowserState -> LarpApplication
handleBrowserEmbed browser request respond = do
    -- TODO: FIXME:...
    let [ValueUInt8 into, ValueUInt32 windowId] = requestArguments request
    if into == 0
        then do
            respond $ Response
                { responseType = ClientError
                , responsePayload = [ValueNothing]
                , responseInternal = Nothing
                }
        else do
            uiReparent (browserUi browser) windowId
            respond $ Response
                { responseType = Ok
                , responsePayload = [ValueNothing]
                , responseInternal = Nothing
                }

{-# WARNING handleTextGet "not implemented" #-}
handleTextGet :: LarpApplication
handleTextGet request respond = do
    case requestHostname request of
        "crowbar.i2p" -> respond $ Response
            { responseType = Ok
            , responsePayload = IF.putResult CategoryText.functionGet [ValueText32 $ TL.pack "Hello, world!\nThis is crowbars fake website!\nGo to someguy's website: lean://someguy.i2p/text/fakepage"]
            , responseInternal = Nothing
            }
        "someguy.i2p" -> respond $ Response
            { responseType = Ok
            , responsePayload = IF.putResult CategoryText.functionGet [ValueText32 $ TL.pack "Someguys's web page: go back to crowbars page: lean://crowbar.i2p/text/no_u"]
            , responseInternal = Nothing
            }
        _             -> respond $ Response
            { responseType = Ok
            , responsePayload = IF.putResult CategoryText.functionGet [ValueText32 $ TL.pack "lean://crowbar.i2p/text/are_you_lost?"]
            , responseInternal = Nothing
            }

handleFunctionNotFound :: LarpApplication
handleFunctionNotFound request respond =
    respond $ Response
        { responseType = ClientError
        , responsePayload = [ValueNothing]
        , responseInternal = Nothing
        }

handleFunctionNotImplemented :: LarpApplication
handleFunctionNotImplemented request respond =
    respond $ Response
        { responseType = ServerError
        , responsePayload = [ValueNothing]
        , responseInternal = Nothing
        }

launchHandler :: String -> String -> String -> IO ()
launchHandler program localport url = do
    -- TODO: FIXME: launch process and wait to check for termination / crashes
    -- TODO: FIXME: don't use spawnCommand, launch process directly
    void $ spawnCommand $ concat [ "cabal new-run ", program, " ", localport, " \"", url, "\"" ]
    --void $ createProcess (proc "cabal" [ "new-run", program, localport, url ])
